from morphapi.api.allenmorphology import AllenMorphology
from vedo import Plotter

am = AllenMorphology()

print(am.neurons.head())

# Select some mouse neurons in the primary visual cortex
neurons = am.neurons.loc[
    (am.neurons.species == "Mus musculus")
    & (am.neurons.structure_area_abbrev == "VISp")
]

# Download some neurons
neurons = am.download_neurons(neurons[:5].id.values)

# ------------------------------- Visualisation ------------------------------ #
print("creating meshes")
neurons = [neuron.create_mesh()[1] for neuron in neurons]

print("visualizing")
vp = Plotter(axes=1)

vp.show(neurons)



import sys

sys.path.append("./")
from morphapi.api.mouselight import MouseLightAPI
from morphapi.api.neuromorphorg import NeuroMorpOrgAPI
from morphapi.api.allenmorphology import AllenMorphology


def test_neuromorpho_download():
    api = NeuroMorpOrgAPI()
    metadata, _ = api.get_neurons_metadata(
        size=2,  # Can get the metadata for up to 500 neurons at the time
        species="mouse",
        cell_type="pyramidal",
        brain_region="neocortex",
    )

    if len(metadata) != 2:
        raise ValueError("Incorrect metadata length")

    neurons = api.download_neurons(metadata)

    if len(neurons) != len(metadata):
        raise ValueError

    neurons = [neuron.create_mesh()[1] for neuron in neurons]


def test_mouselight_download():
    mlapi = MouseLightAPI()

    neurons_metadata = mlapi.fetch_neurons_metadata(
        filterby="soma", filter_regions=["MOs"]
    )

    neurons = mlapi.download_neurons(neurons_metadata[0])

    neurons = [neuron.create_mesh()[1] for neuron in neurons]


def test_allen_morphology_download():
    am = AllenMorphology()

    # Select some mouse neurons in the primary visual cortex
    neurons = am.neurons.loc[
        (am.neurons.species == "Mus musculus")
        & (am.neurons.structure_area_abbrev == "VISp")
    ]

    # Download some neurons
    neurons = am.download_neurons(neurons.sample(3).id.values)

    neurons = [neuron.create_mesh()[1] for neuron in neurons]



import os

from vedo import write, load, Mesh, merge

from morphapi.paths_manager import Paths
from morphapi.utils.data_io import save_yaml, load_yaml


class NeuronCache(Paths):
    cache_filenames_parts = [
        "_soma",
        "_axon",
        "_apical_dendrites",
        "_basal_dendrites",
        "whole_neuron",
    ]

    def __init__(self, **kwargs):
        """
            Initialise API interaction and fetch metadata of neurons in the Allen Database. 
        """
        super().__init__(**kwargs)  # path to data caches

    def get_cache_filenames(self, neuron_name):
        fld = os.path.join(self.meshes_cache, str(neuron_name))
        if not os.path.isdir(fld):
            os.mkdir(fld)
        return [
            os.path.join(fld, str(neuron_name) + part + ".obj")
            for part in self.cache_filenames_parts
        ]

    def get_cache_params_filename(self, neuron_name):
        fld = os.path.join(self.meshes_cache, str(neuron_name))
        if not os.path.isdir(fld):
            os.mkdir(fld)

        return os.path.join(fld, str(neuron_name) + "_params.yml")

    def _check_neuron_mesh_cached(self, neuron_name):
        # If any of the files doesn't exist, the neuron wasn't cached
        for fn in self.get_cache_filenames(neuron_name):
            if not os.path.isfile(fn):
                return False
        return True

    def load_cached_neuron(self, neuron_name, _params):
        if not self._check_neuron_mesh_cached(neuron_name):
            return None

        # Check if params are the same as when cached
        cached_params = load_yaml(self.get_cache_params_filename(neuron_name))
        if len(cached_params) != len(_params):
            changed = cached_params.values()
        else:
            changed = {v for k, v in _params.items() if v != cached_params[k]}
        if changed:
            return None

        # Load neurites
        neurites = [
            "soma",
            "axon",
            "apical_dendrites",
            "basal_dendrites",
            "whole_neuron",
        ]
        loaded = {
            nn: load(fp)
            for nn, fp in zip(neurites, self.get_cache_filenames(neuron_name))
        }

        for nn, act in loaded.items():
            if len(act.points()) == 0:
                loaded[nn] = None

        return loaded

    def write_neuron_to_cache(self, neuron_name, neuron, _params):
        # Write params to file
        save_yaml(self.get_cache_params_filename(neuron_name), _params)

        # Write neurons to file
        file_names = self.get_cache_filenames(neuron_name)

        if isinstance(neuron, Mesh):
            write(neuron, [f for f in file_names if f.endswith("soma.obj")][0])
        else:
            if not isinstance(neuron, dict):
                raise ValueError(
                    f"Invalid neuron argument passed while caching: {neuron}"
                )
            for key, actor in neuron.items():
                if key == "whole_neuron":
                    fname = [f for f in file_names if f.endswith(f"{key}.obj")]
                    write(actor, fname[0])
                else:
                    # Get a single actor for each neuron component.
                    # If there's no data for the component create an empty actor
                    if not isinstance(actor, Mesh):
                        if isinstance(actor, (list, tuple)):
                            if len(actor) == 1:
                                actor = actor[0]
                            elif not actor or actor is None:
                                actor = Mesh()
                            else:
                                try:
                                    actor = merge(actor)
                                except:
                                    raise ValueError(
                                        f"{key} actor should be a mesh or a list of 1 mesh not {actor}"
                                    )

                    if actor is None:
                        actor = Mesh()

                    # Save to file
                    fname = [f for f in file_names if f.endswith(f"{key}.obj")]
                    if fname:
                        write(actor, fname[0])
                    else:
                        raise ValueError(
                            f"No filename found for {key}. Filenames {file_names}"
                        )



import os
from pathlib import Path
from collections import namedtuple
import numpy as np

from vedo.shapes import Sphere, Tube
from vedo import merge
from vedo.colors import colorMap

import neurom as nm
from neurom.core import iter_sections
from neurom.core.dataformat import COLS


from morphapi.morphology.cache import NeuronCache
from morphapi.utils.data_io import get_file_name

import warnings

warnings.filterwarnings("ignore")

component = namedtuple("component", "x y z coords radius component")


class Neuron(NeuronCache):
    _neurite_types = {
        "basal_dendrites": nm.core.types.NeuriteType.basal_dendrite,
        "apical_dendrites": nm.core.types.NeuriteType.apical_dendrite,
        "axon": nm.core.types.NeuriteType.axon,
    }

    _ntypes = nm.core.types.NEURITES

    def __init__(
        self,
        data_file=None,
        neuron_name=None,
        invert_dims=False,
        **kwargs
    ):
        super().__init__(**kwargs)  # path to data caches

        self.invert_dims = invert_dims
        self.neuron_name = neuron_name

        self.data_file = Path(data_file)
        if not self.data_file.exists():
            raise ValueError("The specified path does not exist!")
        self.data_file_type = self.data_file.suffix[1:]

        if self.data_file_type not in ["swc", "json"]:
            raise ValueError("Invalid data file type, should be swc or jon")

        self.load_from_file()

    def load_from_file(self):
        if self.data_file_type is None:
            return
        elif self.data_file_type == "json":
            raise NotImplementedError
        else:
            self.load_from_swc()

    def repair_swc_file(self):
        """ 
            Fixes this: https://github.com/BlueBrain/NeuroM/issues/835
        """
        with open(self.data_file, "r") as read:
            content = read.readlines()

        clean = []
        for line in content:
            if not len(line):
                clean.append(line)
                continue

            line = line.replace("\n", "").replace("\t", " ")
            vals = line.split(" ")
            if len(vals) < 2:
                clean.append(line)
                continue

            if vals[1] != "1" and vals[-1] == "-1":
                vals[-1] = "0"
                clean.append(" ".join(vals))
            else:
                clean.append(line)

        if len(clean) != len(content):
            raise ValueError

        with open(self.data_file, "w") as write:
            for line in clean:
                write.write(f"{line}\n")

    def load_from_swc(self):
        if self.neuron_name is None:
            self.neuron_name = self.data_file.name

        self.repair_swc_file()

        nrn = nm.load_neuron(self.data_file)

        # Get position and radius of some
        soma_pos = nrn.soma.points[0, :3]
        soma_radius = nrn.soma.points[0, -1]

        # Get the rest of the data and store it
        self.morphology = nrn
        self.points = dict(
            soma=component(
                soma_pos[0],
                soma_pos[1],
                soma_pos[2],
                soma_pos,
                soma_radius,
                nrn.soma,
            ),
        )

        for ntype, nclass in self._neurite_types.items():
            self.points[ntype] = [
                component(
                    n.points[:, 0],
                    n.points[:, 1],
                    n.points[:, 2],
                    n.points[:, :3],
                    n.points[:, -1],
                    n,
                )
                for n in nrn.neurites
                if n.type == nclass
            ]

    def _parse_mesh_kwargs(self, **kwargs):
        # To give the entire neuron the same color
        neuron_color = kwargs.pop("neuron_color", None)

        # To give the entire neuron a color based on a cmap
        neuron_number = kwargs.pop("neuron_number", None)
        cmap_lims = kwargs.pop("cmap_lims", (-1, 1))
        cmap = kwargs.pop("cmap", None)

        # To color each component individually
        soma_color = kwargs.pop("soma_color", "salmon")
        apical_dendrites_color = kwargs.pop("apical_dendrites_color", "salmon")
        basal_dendrites_color = kwargs.pop(
            "basal_dendrites_color", apical_dendrites_color
        )
        axon_color = kwargs.pop("axon_color", "salmon")
        whole_neuron_color = kwargs.pop("whole_neuron_color", None)

        # Get each components color from args
        if neuron_color is not None:  # uniform color
            soma_color = (
                apical_dendrites_color
            ) = basal_dendrites_color = axon_color = neuron_color
        elif cmap is not None:  # color according to cmap
            if neuron_number is None:
                neuron_number = 0

            soma_color = colorMap(
                neuron_number, name=cmap, vmin=cmap_lims[0], vmax=cmap_lims[1]
            )
            apical_dendrites_color = (
                basal_dendrites_color
            ) = axon_color = soma_color

        else:  # Use color specified for each component
            pass

        if whole_neuron_color is None:
            whole_neuron_color = soma_color
        return (
            soma_color,
            apical_dendrites_color,
            basal_dendrites_color,
            axon_color,
            whole_neuron_color,
            kwargs,
        )

    def create_mesh(
        self, neurite_radius=2, soma_radius=4, use_cache=True, **kwargs
    ):
        if self.points is None:
            print("No data loaded, returning")
            return

        # Parse kwargs
        (
            soma_color,
            apical_dendrites_color,
            basal_dendrites_color,
            axon_color,
            whole_neuron_color,
            kwargs,
        ) = self._parse_mesh_kwargs(**kwargs)

        if (
            not isinstance(neurite_radius, (int, float))
            or not neurite_radius > 0
        ):
            raise ValueError(
                f"Invalid value for parameter neurite_radius, should be a float > 0"
            )
        if not isinstance(soma_radius, (int, float)) or not soma_radius > 0:
            raise ValueError(
                f"Invalid value for parameter soma_radius, should be a float > 0"
            )
        # prepare params dict for caching
        _params = dict(neurite_radius=neurite_radius, soma_radius=soma_radius)

        # Check if cached files already exist
        if use_cache:
            neurites = self.load_cached_neuron(self.neuron_name, _params)
        else:
            neurites = None

        # Render
        if neurites is not None:
            whole_neuron = neurites.pop("whole_neuron")
            neurites["soma"].c(soma_color)
        else:
            # Create soma actor
            neurites = {}
            if not self.invert_dims:
                coords = self.points["soma"].coords
            else:
                coords = self.points["soma"].coords
                z, y, x = coords[0], coords[1], coords[2]
                coords = np.hstack([x, y, z]).T

            soma = Sphere(
                pos=coords,
                r=self.points["soma"].radius * soma_radius,
                c=soma_color,
            ).computeNormals()
            neurites["soma"] = soma.clone().c(soma_color)

            # Create neurites actors
            for ntype in self._neurite_types:
                actors = []
                for neurite in self.points[ntype]:
                    for section in iter_sections(neurite.component):
                        for child in section.children:
                            if not child.children:
                                coords = child.points[:, COLS.XYZ]
                                if self.invert_dims:
                                    z, y, x = (
                                        coords[:, 0],
                                        coords[:, 1],
                                        coords[:, 2],
                                    )
                                    coords = np.hstack([x, y, z]).T
                                actors.append(Tube(coords, r=neurite_radius))
                            else:
                                for grandchild in child.children:
                                    coords = grandchild.points[:, COLS.XYZ]
                                    if self.invert_dims:
                                        z, y, x = (
                                            coords[:, 0],
                                            coords[:, 1],
                                            coords[:, 2],
                                        )
                                        coords = np.vstack([x, y, z]).T
                                    actors.append(
                                        Tube(coords, r=neurite_radius)
                                    )

                if actors:
                    neurites[ntype] = merge(
                        actors
                    ).computeNormals()  # .smoothMLS2D(f=0.1)
                else:
                    neurites[ntype] = None

            # Merge actors to get the entire neuron
            actors = [
                act.clone() for act in neurites.values() if act is not None
            ]
            whole_neuron = merge(actors).clean().computeNormals()

            # Write to cache
            to_write = neurites.copy()
            to_write["whole_neuron"] = whole_neuron
            self.write_neuron_to_cache(self.neuron_name, to_write, _params)

        # Color actors
        colors = [basal_dendrites_color, apical_dendrites_color, axon_color]
        for n, key in enumerate(
            ["basal_dendrites", "apical_dendrites", "axon"]
        ):
            if neurites[key] is not None:
                neurites[key] = neurites[key].c(colors[n])
        whole_neuron.c(whole_neuron_color)

        return neurites, whole_neuron



from vedo import Plotter

from morphapi.morphology.morphology import Neuron

"""
    This example shows how to use vedo to visualise a 3d reconstruction of a neuron.
    However, the reccomended way to visualise neurons is with brainrender:
    https://github.com/BrancoLab/BrainRender
"""

fp = "examples/example_files/example1.swc"

# Create vedo actors from the .swc file
neuron = Neuron(data_file=fp)
components, neuron = neuron.create_mesh(
    neurite_radius=3,  #
    soma_color="salmon",  # Specify colors [see vedo.colors for more details]
    apical_dendrites_color="darkseagreen",
    basal_dendrites_color="orangered",
    axon_color="blackboard",
    whole_neuron_color="blackboard",
)

# components stores an actor for each neuronal component (dendrites, soma...)
# neuron is a single actor for the entire neuron


# Show neuron as individual components or entire neuron
vp = Plotter(shape=(1, 2), axes=1)
vp.show(*list(components.values()), at=0, interactive=False)
vp.show(neuron, at=1, interactive=True)



import sys

sys.path.append("./")

import os
import pandas as pd
import numpy as np
from tqdm import tqdm

try:
    from allensdk.core.cell_types_cache import CellTypesCache
    from allensdk.api.queries.cell_types_api import CellTypesApi
except ModuleNotFoundError:
    raise ModuleNotFoundError(
        'You need to install the allen sdk package to use AllenMorphology:  "pip install allensdk"'
    )

from morphapi.paths_manager import Paths
from morphapi.utils.data_io import connected_to_internet
from morphapi.morphology.morphology import Neuron


class AllenMorphology(Paths):
    """ Handles the download of neuronal morphology data from the Allen database. """

    def __init__(self, *args, **kwargs):
        """
            Initialise API interaction and fetch metadata of neurons in the Allen Database. 
        """
        if not connected_to_internet():
            raise ConnectionError(
                "You will need to be connected to the internet to use the AllenMorphology class to download neurons"
            )

        Paths.__init__(self, *args, **kwargs)

        # Create a Cache for the Cell Types Cache API
        self.ctc = CellTypesCache(
            manifest_file=os.path.join(
                self.allen_morphology_cache, "manifest.json"
            )
        )

        # Get a list of cell metadata for neurons with reconstructions, download if necessary
        self.neurons = pd.DataFrame(
            self.ctc.get_cells(
                species=[CellTypesApi.MOUSE], require_reconstruction=True
            )
        )
        self.n_neurons = len(self.neurons)

        if not self.n_neurons:
            raise ValueError(
                "Something went wrong and couldn't get neurons metadata from Allen"
            )

        self.downloaded_neurons = self.get_downloaded_neurons()

    def get_downloaded_neurons(self):
        """ 
            Get's the path to files of downloaded neurons
        """
        return [
            os.path.join(self.allen_morphology_cache, f)
            for f in os.listdir(self.allen_morphology_cache)
            if ".swc" in f
        ]

    def download_neurons(self, ids, **kwargs):
        """
            Download neurons and return neuron reconstructions (instances
            of Neuron class)

        :param ids: list of integers with neurons IDs

        """
        if isinstance(ids, np.ndarray):
            ids = list(ids)
        if not isinstance(ids, (list)):
            ids = [ids]

        neurons = []
        print("Downloading neurons")
        for neuron_id in tqdm(ids):
            neuron_file = os.path.join(
                self.allen_morphology_cache, "{}.swc".format(neuron_id)
            )

            # Download file
            self.ctc.get_reconstruction(neuron_id, file_name=neuron_file)

            # Reconstruct neuron
            neurons.append(
                Neuron(neuron_file, neuron_name=str(neuron_id), **kwargs)
            )

        return neurons



import pandas as pd
from pathlib import Path
import zipfile
import shutil
from tqdm import tqdm


from morphapi.paths_manager import Paths
from morphapi.utils.data_io import connected_to_internet
from morphapi.morphology.morphology import Neuron
from bg_space import SpaceConvention
from bg_atlasapi.utils import retrieve_over_http
from bg_atlasapi import BrainGlobeAtlas


def soma_coords_from_file(file_path):
    """Compile dictionary with traced cells origins.
    """
    # Read first line after comment for soma ID:
    line_start = "#"
    with open(file_path, "r") as file:
        while line_start == "#":
            line = file.readline()
            line_start = line[0]

    return [float(p) for p in line.split(" ")[2:-2]]


def fix_mpin_swgfile(file_path, fixed_file_path=None):
    """Fix neurons downloaded from the MPIN website by correcting node
    id and changing the orientation to be standard BrainGlobe.
    """
    if fixed_file_path is None:
        fixed_file_path = file_path

    # Fixed descriptors of the dataset space:
    ORIGIN = "rai"
    SHAPE = [597, 974, 359]
    TARGET_SPACE = "asl"
    NEW_SOMA_SIZE = 7

    bgspace = SpaceConvention(origin=ORIGIN, shape=SHAPE)

    df = pd.read_csv(file_path, sep=" ", header=None, comment="#")

    # In this dataset, soma node is always the first, and
    # other nodes have unspecified identity which we'll set to axon.
    # Hopefully it will be fixed in next iterations of the database.
    df.iloc[0, 1] = 1
    df.iloc[1:, 1] = 2

    # Map points to BrainGlobe orientation:
    df.iloc[:, 2:-2] = bgspace.map_points_to(
        TARGET_SPACE, df.iloc[:, 2:-2].values
    )
    df.iloc[0, -2] = NEW_SOMA_SIZE
    df.to_csv(fixed_file_path, sep=" ", header=None, index=False)


class MpinMorphologyAPI(Paths):
    """Handles the download of neuronal morphology data from the MPIN database.
    """

    def __init__(self, *args, **kwargs):
        Paths.__init__(self, *args, **kwargs)

        self.data_path = Path(self.mpin_morphology) / "fixed"

        if not self.data_path.exists():
            self.download_dataset()

        self._neurons_df = None

    @property
    def neurons_df(self):
        """Table with all neurons positions and soma regions.
        """
        if self._neurons_df is None:
            # Generate table with soma position to query by region:
            atlas = BrainGlobeAtlas("mpin_zfish_1um", print_authors=False)

            neurons_dict = dict()
            for f in self.data_path.glob("*.swc"):
                coords = soma_coords_from_file(f)  # compute coordinates

                # Calculate anatomical structure the neuron belongs to:
                try:
                    region = atlas.structure_from_coords(coords)
                except IndexError:
                    region = 0

                neurons_dict[f.stem] = dict(
                    filename=f.name,
                    pos_ap=coords[0],
                    pos_si=coords[1],
                    pos_lr=coords[2],
                    region=region,
                )

            self._neurons_df = pd.DataFrame(neurons_dict).T

        return self._neurons_df

    def get_neurons_by_structure(self, *region):
        atlas = BrainGlobeAtlas("mpin_zfish_1um", print_authors=False)
        IDs = atlas._get_from_structure(region, "id")
        return list(
            self.neurons_df.loc[self.neurons_df.region.isin(IDs)].index
        )

    def load_neurons(self, neuron_id, **kwargs):
        """
            Load individual neurons given their IDs
        """
        if not isinstance(neuron_id, list):
            neuron_id = [neuron_id]

        to_return = []
        for nid in neuron_id:
            filepath = str(
                Path(self.mpin_morphology)
                / "fixed"
                / self.neurons_df.loc[nid].filename
            )
            to_return.append(
                Neuron(filepath, neuron_name="mpin_" + str(nid), **kwargs,)
            )

        return to_return

    def download_dataset(self):
        """Dowload dataset from Kunst et al 2019.

        """
        if not connected_to_internet():
            raise ValueError(
                "An internet connection is required to download the dataset"
            )
        SOURCE_DATA_DIR = "MPIN-Atlas__Kunst_et_al__neurons_all"

        REMOTE_URL = "https://fishatlas.neuro.mpg.de/neurons/download/download_all_neurons_aligned"

        # # Download folder with all data:
        download_zip_path = Path(self.mpin_morphology) / "data.zip"
        retrieve_over_http(REMOTE_URL, download_zip_path)

        # Uncompress and delete compressed:
        with zipfile.ZipFile(download_zip_path, "r") as zip_ref:
            zip_ref.extractall(download_zip_path.parent)
        download_zip_path.unlink()

        # Fix extracted files:
        extracted_data_path = (
            Path(self.mpin_morphology) / SOURCE_DATA_DIR / "Original"
        )
        self.data_path.mkdir(exist_ok=True)

        for f in tqdm(
            list(extracted_data_path.glob("*.swc")), desc="Fixing swc files"
        ):
            fix_mpin_swgfile(f, self.data_path / f.name)

        shutil.rmtree(extracted_data_path.parent)

        # # 2/1900 neurons still have a little bug, hopefully fixed in the future
        # try:
        #     return Neuron(data_file=fixed_file_path)
        # except:  # Ideally in the next iteration this except won't be necessary
        #     print(f"Unfixable problem while opening {file_path.name}")
        #     return



import sys
from pathlib import Path

"""
    Class to create and store paths to a number of folders uesed to save/load data
"""


# Default paths for Data Folders (store stuff like object meshes, neurons morphology data etc)
default_paths = dict(
    # APIs caches
    allen_morphology_cache="Data/allen_morphology_cache",
    mouselight_cache="Data/mouselight_cache",
    neuromorphorg_cache="Data/neuromorphorg_cache",
    meshes_cache="Data/meshes_cache",
    # Other
    mouse_connectivity_cache="Data/mouse_connectivity_cache",
    mpin_morphology="Data/mpin_morphology"
)


class Paths:

    def __init__(self, base_dir=None, **kwargs):
        """
        Parses a YAML file to get data folders paths. Stores paths to a number of folders used throughtout morphapi. 
        
        :param base_dir: str with path to directory to use to save data. If none the user's base directiry is used. 
        :param kwargs: use the name of a folder as key and a path as argument to specify the path of individual subfolders
        """
        # Get and make base directory

        if base_dir is None:
            self.base_dir = Path.home() / ".morphapi"
        else:
            self.base_dir = base_dir

        self.base_dir.mkdir(exist_ok=True)

        for fld_name, folder in default_paths.items():
            # Check if user provided a path for this folder, otherwise use default

            path = self.base_dir / kwargs.pop(fld_name,
                                              folder)

            # Create folder if it doesn't exist:
            path.mkdir(parents=True, exist_ok=True)
            self.__setattr__(fld_name, str(path))



import os
from tqdm import tqdm

from morphapi.utils.webqueries import request, connected_to_internet
from morphapi.paths_manager import Paths
from morphapi.morphology.morphology import Neuron


class NeuroMorpOrgAPI(Paths):
    _base_url = "http://neuromorpho.org/api/neuron"

    _version = "CNG version"  # which swc version, standardized or original

    def __init__(self, *args, **kwargs):
        if not connected_to_internet():
            raise ConnectionError(
                "You will need to be connected to the internet to use the NeuroMorpOrgAPI class to download neurons"
            )

        Paths.__init__(self, *args, **kwargs)

        # Fields contains the types of fields that can be used to restrict queries
        self.fields = request(self._base_url + "/fields").json()[
            "Neuron Fields"
        ]

    def get_fields_values(self, field):
        """
            Returns the list of allowed values for a given query field
        """
        return list(
            request(self._base_url + f"/fields/{field}").json()["fields"]
        )

    def get_neurons_metadata(self, size=100, page=0, **criteria):
        """
            Uses the neuromorpho API to download metadata about neurons.
            Criteri can be used to restrict the search to neurons of interest/
            http://neuromorpho.org/apiReference.html

            Neuromorpho.org  paginates it's requests so not all neurons metadata
            can be returned at once

            :param size: int in range [0, 500]. Number of neurons whose metadata can be returned at the same time
            :param page: int > 0. Page number. The number of pages depends on size and on how many neurons match the criteria
            :param criteria: use keywords to restrict the query to neurons that match given criteria.
                    keywords should be pass as "field=value". Then only neuron's whose 'field'
                    attribute has value 'value' will be returned.
        """

        if size < 0 or size > 500:
            raise ValueError(
                f"Invalid size argument: {size}. Size should be an integer between 0 and 500"
            )
        if page < 0:
            raise ValueError(
                f"Invalid page argument: {page}. Page should be an integer >= 0"
            )

        url = self._base_url + "/select?q="

        for n, (crit, val) in enumerate(criteria.items()):
            if crit not in self.fields:
                raise ValueError(
                    f"Query criteria {crit} not in available fields: {self.fields}"
                )
            elif val not in self.get_fields_values(crit):
                raise ValueError(
                    f"Query criteria value {val} for field {crit} not valid."
                    + f"Valid values include: {self.get_fields_values(crit)}"
                )

            if isinstance(val, list):
                raise NotImplementedError("Need to join the list")

            if n > 0:
                url += "&"
            url += f"{crit}:{val}"

        url += f"&size={int(size)}&page={int(page)}"
        req = request(url)

        if not req.ok:
            raise ValueError(f"Invalid query with url: {url}")
        else:
            neurons = req.json()

        page = neurons["page"]
        neurons = neurons["_embedded"]["neuronResources"]

        print(
            f"Found metadata for {page['totalElements']} neurons [{page['totalPages']} pages in total]"
            + f"\nReturning metadata about {len(neurons)} neurons from page {page['number']}"
        )

        return neurons, page

    def get_neuron_by_id(self, nid):
        """
            Get a neuron's metadata given it's id number
        """
        return request(self._base_url + f"/id/{nid}").json()

    def get_neuron_by_name(self, nname):
        """
            Get a neuron's metadata given it's name
        """
        return request(self._base_url + f"/name/{nname}").json()

    def download_neurons(self, neurons, _name=None, verbose=True, **kwargs):
        """
            Downloads neuronal morphological data and saves it to .swc files. 
            It then returns a list of Neuron instances with morphological data for each neuron.

            :param neurons: list of neurons metadata (as returned by one of the functions
                        used to fetch metadata)
            :param _name: used internally to save cached neurons with a different prefix when the 
                    class is used to download neurons for other APIs
        """
        if not isinstance(neurons, (list, tuple)):
            neurons = [neurons]

        to_return = []
        if verbose:
            print("Downloading neurons")
        for neuron in tqdm(neurons, disable=not verbose):

            if not isinstance(neuron, dict):
                raise ValueError()

            filepath = os.path.join(
                self.neuromorphorg_cache, f"{neuron['neuron_id']}.swc"
            )

            if not os.path.isfile(filepath):
                # Download and write to file
                if self._version == "CNG version":
                    url = f"http://neuromorpho.org/dableFiles/{neuron['archive'].lower()}/CNG version/{neuron['neuron_name']}.CNG.swc"
                else:
                    url = f"http://neuromorpho.org/dableFiles/{neuron['archive'].lower()}/{self._version}/{neuron['neuron_name']}.swc"

                req = request(url)
                if not req.ok:
                    raise ValueError(
                        f"Failed to getch morphology data for neuron: {neuron['name']}"
                    )

                with open(filepath, "w") as f:
                    f.write(req.content.decode("utf-8"))

            if _name is None:
                _name = "neuromorpho_"

            to_return.append(
                Neuron(
                    filepath,
                    neuron_name=_name + str(neuron["neuron_id"]),
                    **kwargs,
                )
            )

        return to_return



from setuptools import setup, find_namespace_packages
from os import path

this_directory = path.abspath(path.dirname(__file__))
with open(path.join(this_directory, "README.md"), encoding="utf-8") as f:
    long_description = f.read()

requirements = [
    "numpy",
    "pandas",
    "vedo>=2020.3.3",
    "vtk",
    "tqdm",
    "pyyaml>=5.3",
    "neurom",
    "bg_atlasapi",
    "pyinspect",
]

setup(
    name="morphapi",
    version="0.1.1.9",
    description="A lightweight python package to download neuronal morphologies",
    long_description=long_description,
    long_description_content_type="text/markdown",
    install_requires=requirements,
    extras_require={
        "nb": ["jupyter", "k3d"],
        "dev": [
            "pytest-cov",
            "pytest",
            "coveralls",
            "coverage<=4.5.4",
            "pytest-sugar",
            "allensdk",
        ],
    },
    python_requires=">=3.6",
    packages=find_namespace_packages(
        exclude=("Installation", "Meshes", "Metadata", "Screenshots")
    ),
    include_package_data=True,
    url="https://github.com/brainglobe/morphapi",
    author="Federico Claudi",
    zip_safe=False,
)



from tqdm import tqdm
import pandas as pd
from collections import namedtuple

from bg_atlasapi import BrainGlobeAtlas

from morphapi.utils.webqueries import post_mouselight, mouselight_base_url
from morphapi.paths_manager import Paths
from morphapi.utils.data_io import is_any_item_in_list, flatten_list
from morphapi.api.neuromorphorg import NeuroMorpOrgAPI


"""
    Collections of functions to query http://ml-neuronbrowser.janelia.org/ and get data about either the status of the API, 
    the brain regions or the neurons available. 
    Queries are sent by sending POST requests to http://ml-neuronbrowser.janelia.org/graphql
    with a string query. 
"""

# ---------------------------------------------------------------------------- #
#                                  QUERY UTILS                                 #
# ---------------------------------------------------------------------------- #


def mouselight_api_info():
    """
        Get the number of cells available in the database
    """
    # Get info from the ML API
    url = mouselight_base_url + "graphql"

    query = """
            query {
                queryData {
                    totalCount
                    }
                }
            """
    res = post_mouselight(url, query=query)
    print(
        "{} neurons on MouseLight database. ".format(
            res["queryData"]["totalCount"]
        )
    )


def mouselight_get_brainregions():
    """
        Get metadata about the brain brain regions as they are known by Janelia's Mouse Light. 
        IDs and Names sometimes differ from Allen's CCF.
    """

    # Download metadata about brain regions from the ML API
    url = mouselight_base_url + "graphql"
    # query =  "systemSettings {apiVersion apiRelease neuronCount\}}"
    query = """
            query {
                brainAreas{
                    
                    acronym
                    name
                    id
                    atlasId
                    graphOrder
                    parentStructureId
                    structureIdPath
                }
            }
            """
    res = post_mouselight(url, query=query)["brainAreas"]

    # Clean up and turn into a dataframe
    keys = {k: [] for k in res[0].keys()}
    for r in res:
        for k in r.keys():
            keys[k].append(r[k])

    structures_data = pd.DataFrame.from_dict(keys)
    return structures_data


def mouselight_structures_identifiers():
    """
    When the data are downloaded as SWC, each node has a structure identifier ID to tell if it's soma, axon or dendrite.
    This function returns the ID number --> structure table. 
    """

    # Download the identifiers used in ML neurons tracers
    url = mouselight_base_url + "graphql"
    # query =  "systemSettings {apiVersion apiRelease neuronCount\}}"
    query = """
            query {
                structureIdentifiers{
                    id
                    name
                    value
                }
            }
        """
    res = post_mouselight(url, query=query)["structureIdentifiers"]

    keys = {k: [] for k in res[0].keys()}
    for r in res:
        for k in r.keys():
            keys[k].append(r[k])

    structures_identifiers = pd.DataFrame.from_dict(keys)
    return structures_identifiers


def make_query(filterby=None, filter_regions=None, invert=False):
    """
    Constructs the strings used to submit graphql queries to the mouse light api

    :param filterby: str, soma, axon on dendrite. Search by neurite structure (Default value = None)
    :param filter_regions:  list, tuple. list of strings. Acronyms of brain regions to use for query (Default value = None)
    :param invert:  If true the inverse of the query is return (i.e. the neurons NOT in a brain region) (Default value = False)

    """
    searchneurons = """
                queryTime
                totalCount
                
                neurons{
                tag
                id
                idNumber
                idString
                
                brainArea{
                    id
                    acronym
                    name
                    safeName
                    atlasId
                    aliases
                    structureIdPath
                }

                tracings{
                    soma{
                    x
                    y
                    z
                    radius
                    brainArea{
                        id
                        acronym
                    }
                    sampleNumber
                    parentNumber
                    
                    }
                
                id
                tracingStructure{
                    name
                    value
                    id
                }
                }
            }
    """

    if filterby is None or filterby == "soma":
        query = """
                    query {{
                        searchNeurons {{
                            {}
                        }}
                    }}
                    """.format(
            searchneurons
        )
    else:
        raise NotImplementedError("This feature is not available yet")
        # Get predicate type
        if filterby.lower() in ["axon", "axons", "end point", "branch point"]:
            predicateType = 1

        elif filterby.lower() in [
            "dendrite",
            "apical dendrite",
            "(basal) dendrite",
        ]:
            raise NotImplementedError
            filterby = "(basal) dendrite"
            predicateType = 2
        else:
            raise ValueError("invalid search by argument")

        # Get neuron structure id
        structures_identifiers = mouselight_structures_identifiers()
        structureid = str(
            structures_identifiers.loc[
                structures_identifiers.name == filterby
            ]["id"].values[0]
        )

        # Get brain regions ids
        brainregions = mouselight_get_brainregions()
        brainareaids = [
            str(brainregions.loc[brainregions.acronym == a]["id"].values[0])
            for a in filter_regions
        ]

        # Get inversion
        if invert:
            invert = "true"
        else:
            invert = "false"

        query = """
        query {{
            searchNeurons (
                context: {{
                scope: 6
                predicates: [{{
                    predicateType: {predicate}
                    tracingIdsOrDOIs: []
                    tracingIdsOrDOIsExactMatch: false
                    tracingStructureIds: []
                    amount: 0
                    nodeStructureIds: ['{structure}']
                    brainAreaIds: {brainarea}
                    invert: {invert}
                    composition: 1
                    }}]
                }}
            ) {{
                {base}
            }}
        }}
        """.format(
            predicate=predicateType,
            structure=str(structureid),
            brainarea=brainareaids,
            invert=invert,
            base=searchneurons,
        )

        query = query.replace("\\t", "").replace("'", '"')
    return query


# ---------------------------------------------------------------------------- #
#                                  MAIN CLASS                                  #
# ---------------------------------------------------------------------------- #


class MouseLightAPI(Paths):
    def __init__(self, base_dir=None, **kwargs):
        """
            Handles the download of neurons morphology data from the Mouse Light project

            :param base_dir: path to directory to use for saving data (default value None)
            :param kwargs: can be used to pass path to individual data folders. See morphapi/utils /paths_manager.py
        """
        Paths.__init__(self, base_dir=base_dir, **kwargs)

    def fetch_neurons_metadata(
        self, filterby=None, filter_regions=None, **kwargs
    ):
        """
        Download neurons metadata and data from the API. The downloaded metadata can be filtered to keep only
        the neurons whose soma is in a list of user selected brain regions.
        
        :param filterby: Accepted values: "soma". If it's "soma", neurons are kept only when their soma
                        is in the list of brain regions defined by filter_regions (Default value = None)
        :param filter_regions: List of brain regions acronyms. If filtering neurons, these specify the filter criteria. (Default value = None)
        :param **kwargs: 

        """
        # Download all metadata
        print("Querying MouseLight API...")
        url = mouselight_base_url + "graphql"
        query = make_query(
            filterby=filterby, filter_regions=filter_regions, **kwargs
        )

        res = post_mouselight(url, query=query)["searchNeurons"]
        print(
            "     ... fetched metadata for {} neurons in {}s".format(
                res["totalCount"], round(res["queryTime"] / 1000, 2)
            )
        )

        # Process neurons to clean up the results and make them easier to handle
        neurons = res["neurons"]
        node = namedtuple("node", "x y z r area_acronym sample_n parent_n")
        tracing_structure = namedtuple(
            "tracing_structure", "id name value named_id"
        )

        cleaned_nurons = []  # <- output is stored here
        for neuron in neurons:
            if neuron["brainArea"] is not None:
                brainArea_acronym = neuron["brainArea"]["acronym"]
                brainArea_id = neuron["brainArea"]["id"]
                brainArea_name = neuron["brainArea"]["name"]
                brainArea_safename = neuron["brainArea"]["safeName"]
                brainArea_atlasId = neuron["brainArea"]["atlasId"]
                brainArea_aliases = neuron["brainArea"]["aliases"]
                brainArea_structureIdPath = neuron["brainArea"][
                    "structureIdPath"
                ]
            else:
                brainArea_acronym = None
                brainArea_id = None
                brainArea_name = None
                brainArea_safename = None
                brainArea_atlasId = None
                brainArea_aliases = None
                brainArea_structureIdPath = None

            if len(neuron["tracings"]) > 1:
                dendrite = tracing_structure(
                    neuron["tracings"][1]["id"],
                    neuron["tracings"][1]["tracingStructure"]["name"],
                    neuron["tracings"][1]["tracingStructure"]["value"],
                    neuron["tracings"][1]["tracingStructure"]["id"],
                )
            else:
                dendrite = None

            clean_neuron = dict(
                brainArea_acronym=brainArea_acronym,
                brainArea_id=brainArea_id,
                brainArea_name=brainArea_name,
                brainArea_safename=brainArea_safename,
                brainArea_atlasId=brainArea_atlasId,
                brainArea_aliases=brainArea_aliases,
                brainArea_structureIdPath=brainArea_structureIdPath,
                id=neuron["id"],
                idNumber=neuron["idNumber"],
                idString=neuron["idString"],
                tag=neuron["tag"],
                soma=node(
                    neuron["tracings"][0]["soma"]["x"],
                    neuron["tracings"][0]["soma"]["y"],
                    neuron["tracings"][0]["soma"]["z"],
                    neuron["tracings"][0]["soma"]["radius"],
                    neuron["tracings"][0]["soma"]["brainArea"],
                    neuron["tracings"][0]["soma"]["sampleNumber"],
                    neuron["tracings"][0]["soma"]["parentNumber"],
                ),
                axon=tracing_structure(
                    neuron["tracings"][0]["id"],
                    neuron["tracings"][0]["tracingStructure"]["name"],
                    neuron["tracings"][0]["tracingStructure"]["value"],
                    neuron["tracings"][0]["tracingStructure"]["id"],
                ),
                dendrite=dendrite,
            )
            cleaned_nurons.append(clean_neuron)

        # Filter neurons to keep only those matching the search criteria
        if filterby is not None:
            if filter_regions is None:
                raise ValueError(
                    "If filtering neuron by region, you need to pass a list of filter regions to use"
                )

            # get brain globe atlas
            atlas = BrainGlobeAtlas("allen_mouse_25um")

            # Filter by soma
            if filterby == "soma":
                filtered_neurons = []
                for neuron in cleaned_nurons:
                    if neuron["brainArea_acronym"] is None:
                        continue

                    # get ancestors of neuron's regions
                    try:
                        neuron_region_ancestors = atlas.get_structure_ancestors(
                            neuron["brainArea_acronym"]
                        )
                    except KeyError:
                        # ignore if region is not found
                        continue

                    # If any of the ancestors are in the allowed regions, keep neuron.
                    if is_any_item_in_list(
                        filter_regions, neuron_region_ancestors
                    ):
                        filtered_neurons.append(neuron)
                print(
                    "	... selected {} neurons out of {}".format(
                        len(filtered_neurons), res["totalCount"]
                    )
                )

                neurons = filtered_neurons
            else:
                print(
                    "	... selected {} neurons out of {}".format(
                        len(cleaned_nurons), res["totalCount"]
                    )
                )
                neurons = cleaned_nurons
        else:
            neurons = cleaned_nurons

        return neurons

    def download_neurons(self, neurons_metadata, **kwargs):
        """
        Given a list of neurons metadata from self.fetch_neurons_metadata
        this funcition downloads the morphological data.
        The data are actually downloaded from neuromorpho.org
        
        :param neurons_metadata: list with metadata for neurons to download
        :returns: list of Neuron instances

        """
        if not isinstance(neurons_metadata, (list, tuple)):
            neurons_metadata = [neurons_metadata]
        nmapi = NeuroMorpOrgAPI()
        nmapi._version = "Source-Version"

        neurons = []
        print("Downloading neurons")
        for neuron in tqdm(neurons_metadata):

            neurons.append(
                nmapi.download_neurons(
                    nmapi.get_neuron_by_name(neuron["idString"]),
                    _name="mouselight_",
                    invert_dims=True,
                    verbose=False,
                )
            )

        # ? Old code
        # def get(url, tracing_id): # send a query for a single tracing ID
        #     """
        #     Creates the URL for each neuron to download

        #     :param url: str with url
        #     :param tracing_id: str with the neuron's ID

        #     """
        #     query = {"ids":[tracing_id]}
        #     res = post_mouselight(url, query=query, clean=True)['tracings']
        #     return res

        # if neurons_metadata is None or not neurons_metadata:
        #     return None
        # print("Downloading neurons tracing data from Mouse Light")

        # # check arguments
        # if not isinstance(neurons_metadata, list): neurons_metadata = [neurons_metadata]

        # # URL used to fetch neurons
        # url = mouselight_base_url + "tracings/tracings"

        # # loop over neurons
        # neurons = []
        # for neuron in tqdm(neurons_metadata):
        #     # Check if a .swc file already exists for this neuron
        #     file_path = os.path.join(self.mouselight_cache, neuron['idString']+".swc")

        #     if not os.path.isfile(file_path):
        #         # Download and save as swc
        #         # Get tracings by requests
        #         axon_tracing, dendrite_tracing = None, None
        #         if neuron['axon'] is not None:
        #             axon_tracing = get(url, neuron['axon'].id)
        #         if neuron['dendrite'] is not None:
        #             dendrite_tracing = get(url, neuron['dendrite'].id)

        #         self.make_swc(neuron, file_path, axon_tracing=axon_tracing, dendrite_tracing=dendrite_tracing)

        #     # Reconstruct from .swc
        #     neurons.append(Neuron(file_path,))

        return flatten_list(neurons)


"""
Old code to fetch morphological data from mouselight database
"""
# @staticmethod
# def make_json(neuron, file_path,  axon_tracing=None, dendrite_tracing=None):
#     """
#     Creates a .json file with the neuron's data that can be read by the mouselight parser.

#     :param neuron: dict with neuron's data
#     :param file_path: str, path where to save the json file
#     :param axon_tracing: list with data for axon tracing (Default value = None)
#     :param dendrite_tracing: list with data for dendrite tracing (Default value = None)

#     """
#     # parse axon
#     if axon_tracing is not None:
#         nodes = axon_tracing[0]['nodes']
#         axon = [
#             dict(
#                 sampleNumber = n['sampleNumber'],
#                 x = n['x'],
#                 y = n['y'],
#                 z = n['z'],
#                 radius = n['radius'],
#                 parentNumber = n['parentNumber'],
#             )
#             for n in nodes]
#     else:
#         axon = []

#     # parse dendrites
#     if dendrite_tracing is not None:
#         nodes = dendrite_tracing[0]['nodes']
#         dendrite = [
#             dict(
#                 sampleNumber = n['sampleNumber'],
#                 x = n['x'],
#                 y = n['y'],
#                 z = n['z'],
#                 radius = n['radius'],
#                 parentNumber = n['parentNumber'],
#             )
#             for n in nodes]
#     else:
#         dendrite = []

#     content = dict(
#         neurons = [
#             dict(
#                 idString = neuron['idString'],
#                 soma = dict(
#                     x = neuron['soma'].x,
#                     y = neuron['soma'].y,
#                     z = neuron['soma'].z,
#                 ),
#                 axon = axon,
#                 dendrite = dendrite,

#             )
#         ]
#     )


#     # save to file
#     if not filepath.endswith('.json'):
#         raise ValueError(f'Invalid file path {filepath}')
#     with open(file_path, 'w') as f:
#         json.dump(content, f)

# @staticmethod
# def make_swc(neuron, file_path,  axon_tracing=None, dendrite_tracing=None):
#     """
#     Creates a .swc file with the neuron's data that can be read by the mouselight parser.

#     :param neuron: dict with neuron's data
#     :param file_path: str, path where to save the swc file
#     :param axon_tracing: list with data for axon tracing (Default value = None)
#     :param dendrite_tracing: list with data for dendrite tracing (Default value = None)

#     """

#     if not file_path.endswith('.swc'):
#         raise ValueError(f'Invalid file path {file_path}')

#     # parse data
#     if axon_tracing is not None:
#         nodes = axon_tracing[0]['nodes']
#         axon = [
#             dict(
#                 sampleNumber = n['sampleNumber'],
#                 x = n['x'],
#                 y = n['y'],
#                 z = n['z'],
#                 id = 2,
#                 radius = n['radius'],
#                 parentNumber = n['parentNumber'],
#             )
#             for n in nodes]
#     else:
#         axon = []

#     if dendrite_tracing is not None:
#         nodes = dendrite_tracing[0]['nodes']
#         dendrite = [
#             dict(
#                 sampleNumber = n['sampleNumber'],
#                 x = n['x'],
#                 y = n['y'],
#                 z = n['z'],
#                 radius = n['radius'],
#                 id = 3,
#                 parentNumber = n['parentNumber'],
#             )
#             for n in nodes]
#     else:
#         dendrite = []

#     # Save as swc
#     with open(file_path, 'w') as out:
#         out.write(f"# neuron: {neuron['idString']} downloaded from mouselight database [janelia campus]\n")
#         out.write("# .swc reconstructed using morphapi python package https://github.com/brainglobe/morphapi\n\n")
#         out.write("# id,type,x,y,z,r,pid\n")
#         out.write(f" 1 1 {neuron['soma'].x} "+
#                     f"{neuron['soma'].y} {neuron['soma'].z} "+
#                     f"{neuron['soma'].r} -1\n")


#         neurites = axon + dendrite
#         sort_idx = np.argsort([n['parentNumber'] for n in neurites])
#         neurites = list(np.array(neurites)[sort_idx])

#         for n in neurites:
#             if n['parentNumber'] == -1:
#                 pn = 1
#             else:
#                 pn = n['parentNumber']
#             out.write(f" {n['sampleNumber']} {n['id']} {n['x']} {n['y']} {n['z']} {n['radius']} {pn}\n")



from vedo import Plotter

from morphapi.api.mouselight import MouseLightAPI


# ---------------------------- Downloading neurons --------------------------- #
mlapi = MouseLightAPI()

# Fetch metadata for neurons with soma in the secondary motor cortex
neurons_metadata = mlapi.fetch_neurons_metadata(
    filterby="soma", filter_regions=["MOs"]
)

# Then we can download the files and save them as a .json file
neurons = mlapi.download_neurons(neurons_metadata[0])


"""
    mlapi.download_neurons returns a list of instances of the class Neuron
    from morphapi.morphology.morphology.
"""


# ------------------------------- Visualisation ------------------------------ #
print("creating meshes")
neurons = [neuron.create_mesh()[1] for neuron in neurons]

print("visualizing")
vp = Plotter(shape=(1, len(neurons)), axes=1)

vp.show(neurons)
